from ._summary import FailureSummary, SuccessSummary
from .exceptions import FailureError


class Call(object):
    def got_failure(self, ctx, method_name, reason):
        raise FailureError(reason)

    def got_result(self, value):
        return value

    def finished(self):
        pass


class Run(object):
    def __init__(self, protocol):
        self.protocol = protocol

    def got_failure(self, ctx, method_name, reason):
        return FailureSummary(self.protocol, ctx, method_name, reason)

    def got_result(self, value):
        return SuccessSummary(self.protocol, value)

    def finished(self):
        return SuccessSummary(self.protocol, None)



from .._context import assign_namespace
from .._marker import BeginningOfStory, EndOfStory
from .._return import Failure, Result, Skip, Success


def execute(runner, ctx, history, methods):

    skipped = 0

    for method, contract, protocol in methods:

        method_type = type(method)

        if skipped > 0:
            if method_type is EndOfStory:
                skipped -= 1
            elif method_type is BeginningOfStory:
                skipped += 1
            continue

        history.before_call(method.__name__)

        try:
            result = method(ctx)
        except Exception as error:
            history.on_error(error.__class__.__name__)
            raise

        restype = type(result)
        assert restype in (Result, Success, Failure, Skip)

        if restype is Failure:
            try:
                protocol.check_return_statement(method, result.reason)
            except Exception as error:
                history.on_error(error.__class__.__name__)
                raise
            history.on_failure(result.reason)
            return runner.got_failure(ctx, method.__name__, result.reason)

        if restype is Result:
            history.on_result(result.value)
            return runner.got_result(result.value)

        if restype is Skip:
            history.on_skip()
            skipped = 1
            continue

        if method_type is BeginningOfStory:
            try:
                contract.check_story_arguments(ctx)
            except Exception as error:
                history.on_error(error.__class__.__name__)
                raise
            history.on_substory_start()
            continue

        if method_type is EndOfStory:
            history.on_substory_end()
            continue

        try:
            contract.check_success_statement(method, ctx, result.kwargs)
        except Exception as error:
            history.on_error(error.__class__.__name__)
            raise

        assign_namespace(ctx, method, result.kwargs)

    return runner.finished()



from ._argument import get_arguments
from ._collect import collect_story
from ._failures import check_data_type
from ._mounted import ClassMountedStory, MountedStory
from ._wrap import wrap_story


class Story(object):
    def __init__(self, f):
        self.name = f.__name__
        self.arguments = get_arguments(f)
        self.collected = collect_story(f)
        self.failures(None)

    def __get__(self, obj, cls):
        if obj is None:
            return ClassMountedStory(cls, self.name, self.collected, self.failures)
        else:
            methods, failures = wrap_story(
                self.arguments,
                self.collected,
                cls.__name__,
                self.name,
                obj,
                self.__failures,
            )
            return MountedStory(
                obj, cls.__name__, self.name, self.arguments, methods, failures
            )

    def failures(self, failures):
        check_data_type(failures)
        self.__failures = failures
        return failures



from ._return import Success


class BeginningOfStory(object):
    def __init__(self, cls_name, name):
        self.cls_name = cls_name
        self.name = name
        self.parent_name = None
        self.same_object = None

    def __call__(self, ctx):
        return Success()

    @property
    def __name__(self):
        if self.parent_name is None:
            return self.cls_name + "." + self.name
        elif self.same_object:
            return self.parent_name
        else:
            return self.parent_name + " (" + self.cls_name + "." + self.name + ")"

    def set_parent(self, parent_name, same_object):
        self.parent_name = parent_name
        self.same_object = same_object


class EndOfStory(object):
    def __init__(self, is_empty):
        self.is_empty = is_empty

    def __call__(self, ctx):
        return Success()

    __name__ = "end_of_story"



from .exceptions import ContextContractError


def make_contract(cls_name, name, arguments):
    return Contract(cls_name, name, arguments)


def validate_arguments(arguments, args, kwargs):
    # FIXME: Should be a method of the `Contract` class.
    assert not (args and kwargs)

    if args:
        assert len(arguments) == len(args)
        return [(k, v) for k, v in zip(arguments, args)]

    assert set(arguments) == set(kwargs)
    return [(k, kwargs[k]) for k in arguments]


class Contract(object):
    def __init__(self, cls_name, name, arguments):
        self.cls_name = cls_name
        self.name = name
        self.arguments = arguments

    def check_story_arguments(self, ctx):
        missed = set(self.arguments) - set(ctx._Context__ns)
        if missed:
            message = missed_variable_template.format(
                missed=", ".join(sorted(missed)),
                cls=self.cls_name,
                method=self.name,
                arguments=", ".join(self.arguments),
                ctx=ctx,
            )
            raise ContextContractError(message)

    def check_success_statement(self, method, ctx, ns):
        tries_to_override = set(ctx._Context__ns) & set(ns)
        if tries_to_override:
            message = variable_override_template.format(
                variables=", ".join(map(repr, sorted(tries_to_override))),
                cls=method.__self__.__class__.__name__,
                method=method.__name__,
            )
            raise ContextContractError(message)


def deny_attribute_assign():
    raise ContextContractError(assign_attribute_template)


def deny_attribute_delete():
    raise ContextContractError(delete_attribute_template)


missed_variable_template = """
These variables are missing from the context: {missed}

Story method: {cls}.{method}

Story arguments: {arguments}

{ctx!r}
""".strip()


variable_override_template = """
These variables are already present in the context: {variables}

Function returned value: {cls}.{method}

Use different names for Success() keyword arguments.
""".strip()


assign_attribute_template = """
Context object is immutable.

Use Success() keyword arguments to expand its scope.
""".strip()


delete_attribute_template = """
Context object is immutable.

Variables can not be removed from Context.
""".strip()



"""
stories
-------

This module implements Business Transaction DSL.

:copyright: (c) 2018-2019 dry-python team.
:license: BSD, see LICENSE for more details.
"""

from ._argument import arguments
from ._return import Failure, Result, Skip, Success
from ._story import Story as story


__all__ = ["story", "arguments", "Result", "Success", "Failure", "Skip"]



"""
stories.shortcuts
-----------------

This module contains convenient functions to reduce boilerplate code.

:copyright: (c) 2018-2019 dry-python team.
:license: BSD, see LICENSE for more details.
"""

from ._mounted import ClassMountedStory


def failures_in(cls, *args):
    def setter(failures):
        for attrname in dir(cls):
            attribute = getattr(cls, attrname)
            if type(attribute) is ClassMountedStory:
                attribute.failures(failures)
        return failures

    if args:
        return setter(*args)
    else:
        return setter



"""
stories.contrib.sentry.breadcrumbs
----------------------------------

This module contains integration with Sentry.

:copyright: (c) 2018-2019 dry-python team.
:license: BSD, see LICENSE for more details.
"""

import stories._context
from raven.breadcrumbs import libraryhook, record


origin_context_init = stories._context.Context.__init__


@libraryhook("stories")
def track_context():
    def wrapper(ctx, ns, history):
        origin_context_init(ctx, ns, history)
        record(
            processor=lambda data: data.update(
                {"category": "story", "message": repr(ctx)}
            )
        )

    stories._context.Context.__init__ = wrapper



from collections import OrderedDict

from ._contract import deny_attribute_assign, deny_attribute_delete


class Context(object):
    def __init__(self, ns, history):
        self.__dict__["_Context__ns"] = OrderedDict(ns)
        self.__dict__["_Context__history"] = history
        self.__dict__["_Context__lines"] = ["Story argument"] * len(ns)

    def __getattr__(self, name):
        return self.__ns[name]

    def __setattr__(self, name, value):
        deny_attribute_assign()

    def __delattr__(self, name):
        deny_attribute_delete()

    def __repr__(self):
        return history_representation(self) + "\n\n" + context_representation(self)

    def __dir__(self):
        spec = type("Context", (object,), {})
        parent = set(dir(spec()))
        current = set(self.__dict__) - {
            "_Context__ns",
            "_Context__history",
            "_Context__lines",
        }
        scope = set(self.__ns)
        attributes = sorted(parent | current | scope)
        return attributes


def assign_namespace(ctx, method, kwargs):
    ctx._Context__ns.update(kwargs)
    line = "Set by %s.%s" % (method.__self__.__class__.__name__, method.__name__)
    ctx._Context__lines.extend([line] * len(kwargs))


def history_representation(ctx):
    result = "\n".join(ctx._Context__history.lines)
    return result


def context_representation(ctx):
    if not ctx._Context__lines:
        return "Context()"
    items = [
        "%s = %s" % (key, repr(value)) for (key, value) in ctx._Context__ns.items()
    ]
    longest = max(map(len, items))
    lines = [
        "    %s  # %s" % (item.ljust(longest), line)
        for item, line in zip(items, ctx._Context__lines)
    ]
    return "\n".join(["Context:"] + lines)



"""
stories.contrib.pytest
----------------------

This module contains integration with PyTest framework.

:copyright: (c) 2018-2019 dry-python team.
:license: BSD, see LICENSE for more details.
"""

import linecache
import sys
import textwrap

from _pytest.config import hookimpl

import stories._context


origin_context_init = stories._context.Context.__init__


def track_context(storage):
    def wrapper(ctx, ns, history):
        origin_context_init(ctx, ns, history)
        storage.append((get_test_source(*get_test_call()), ctx))

    return wrapper


def get_test_call():
    f = sys._getframe()

    while True:
        if (
            "@py_builtins" in f.f_globals
            and "@pytest_ar" in f.f_globals
            and f.f_code.co_filename != __file__
        ):
            return f.f_code.co_filename, f.f_lineno
        elif not f.f_back:
            raise Exception("Can not find running test")
        else:
            f = f.f_back


def get_test_source(filename, lineno):

    start = max(1, lineno - 3)
    end = lineno + 3
    adjust_to = len(str(end))

    lines = [linecache.getline(filename, no) for no in range(start, end)]
    text = textwrap.dedent("".join(lines))

    src = []
    for num, line in zip(range(start, end), text.splitlines()):
        sep = "->" if num == lineno else "  "
        src.append((" %s %s %s" % (str(num).rjust(adjust_to), sep, line)).rstrip())
    src = "\n".join(src)

    return src


@hookimpl(hookwrapper=True)
def pytest_runtest_call(item):
    storage = []
    stories._context.Context.__init__ = track_context(storage)
    yield
    stories._context.Context.__init__ = origin_context_init
    for i, (src, ctx) in enumerate(storage, 1):
        output = "\n\n".join([src, repr(ctx)])
        item.add_report_section("call", "story #%d" % (i,), output)



# -*- coding: utf-8 -*-
from distutils.core import setup

package_dir = \
{'': 'src'}

packages = \
['stories',
 'stories._exec',
 'stories.contrib',
 'stories.contrib.debug_toolbars',
 'stories.contrib.debug_toolbars.django',
 'stories.contrib.sentry']

package_data = \
{'': ['*'],
 'stories.contrib.debug_toolbars.django': ['templates/stories/debug_toolbar/*']}

entry_points = \
{'pytest11': ['stories = stories.contrib.pytest']}

setup_kwargs = {
    'name': 'stories',
    'version': '0.10.1',
    'description': 'Define a user story in the business transaction DSL',
    'long_description': ".. |travis| image:: https://travis-ci.org/dry-python/stories.svg?branch=master\n    :target: https://travis-ci.org/dry-python/stories\n\n.. |codecov| image:: https://codecov.io/gh/dry-python/stories/branch/master/graph/badge.svg\n    :target: https://codecov.io/gh/dry-python/stories\n\n.. |docs| image:: https://readthedocs.org/projects/stories/badge/?version=latest\n    :target: https://stories.readthedocs.io/en/latest/?badge=latest\n\n.. |gitter| image:: https://badges.gitter.im/dry-python/stories.svg\n    :target: https://gitter.im/dry-python/stories\n\n.. |pypi| image:: https://img.shields.io/pypi/v/stories.svg\n    :target: https://pypi.python.org/pypi/stories/\n\n.. |black| image:: https://img.shields.io/badge/code%20style-black-000000.svg\n    :target: https://github.com/ambv/black\n\n.. image:: https://raw.githubusercontent.com/dry-python/brand/master/logo/stories.png\n\n|travis| |codecov| |docs| |gitter| |pypi| |black|\n\n----\n\nThe business transaction DSL\n============================\n\n- `Source Code`_\n- `Issue Tracker`_\n- `Documentation`_\n- `Discussion`_\n\nInstallation\n------------\n\nAll released versions are hosted on the Python Package Index.  You can\ninstall this package with following command.\n\n.. code:: bash\n\n    pip install stories\n\nUsage\n-----\n\n``stories`` provide a simple way to define a complex business scenario\nthat include many processing steps.\n\n.. code:: python\n\n    from stories import story, arguments, Success, Failure, Result\n\n    class Subscribe:\n\n        @story\n        @arguments('category_id', 'user_id')\n        def buy(I):\n\n            I.find_category\n            I.find_profile\n            I.check_balance\n            I.persist_subscription\n            I.show_subscription\n\n        def find_category(self, ctx):\n\n            category = Category.objects.get(id=ctx.category_id)\n            return Success(category=category)\n\n        def find_profile(self, ctx):\n\n            profile = Profile.objects.get(user_id=ctx.user_id)\n            return Success(profile=profile)\n\n        def check_balance(self, ctx):\n\n            if ctx.category.cost < ctx.profile.balance:\n                return Success()\n            else:\n                return Failure()\n\n        def persist_subscription(self, ctx):\n\n            subscription = Subscription(ctx.category, ctx.profile)\n            subscription.save()\n            return Success(subscription=subscription)\n\n        def show_subscription(self, ctx):\n\n            return Result(ctx.subscription)\n\n.. code:: python\n\n    >>> Subscribe().buy(category_id=1, user_id=1)\n    <Subscription object>\n    >>> _\n\nThis code style allow you clearly separate actual business scenario\nfrom implementation details.\n\nLicense\n-------\n\nStories library is offered under the two clause BSD license.\n\n.. _source code: https://github.com/dry-python/stories\n.. _issue tracker: https://github.com/dry-python/stories/issues\n.. _documentation: https://stories.readthedocs.io/en/latest/\n.. _discussion: https://gitter.im/dry-python/stories\n",
    'author': 'Artem Malyshev',
    'author_email': 'proofit404@gmail.com',
    'url': 'https://dry-python.org/',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'entry_points': entry_points,
    'python_requires': '>=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*',
}


setup(**setup_kwargs)
